%**************************************************************************
% Spectral method: simulations & predictions
% CDP model

% By Junjie Ma
%**************************************************************************
tic
close all
clear all;
clc

randn('state',0);
rand('state',0);

namestr = 'venere' ;  % stanford, galaxy, columbia
stanstr = 'png';
Mask_opt = 'Y';

SE_only = 'N';

X_full  = mat2gray(imread([namestr,'.',stanstr])) ;

X_full       = imresize(X_full, 0.1);
NSIM = 2;

X_tmp = squeeze(X_full(:,:,1)); % Image x is n1 x n2 

n1      = size(X_full,1);
n2      = size(X_full,2);

N = n1 * n2;


Methods = {'optimal_Lu','optimal','subset','truncation'};
% Methods = {'subset'};

Lu_eps = 0.1;

alpha_y = 2;

alpha_y_subset = 1.5;

delta_grid = [3 4 5 6 7 8];
% % delta_grid = 4;
Iteration_power = 10000;


  
Alpha_power = zeros(length(delta_grid),length(Methods));  
Conv_error = zeros(length(delta_grid),length(Methods));

Alpha_SE = zeros(length(delta_grid),length(Methods));
Rho_x2 = zeros(length(delta_grid),length(Methods));

t1 = clock;

for mmm = 1:length(Methods)
    
    Trimming_opt = Methods{mmm};
    
    for ddd = 1:length(delta_grid)

        delta = delta_grid(ddd);
        
        L = delta;

        M = fix(delta * N);                        % number of measurements
        y = zeros(n1,n2,3,L);
        Masks = zeros(n1,n2,L);

        %% SE
        % search mu;

        flag = 1;
        MM = 1e7;
        Z = sqrt(1/delta) * (randn(MM,1)+1i*randn(MM,1))/sqrt(2);
        Y = abs(Z);
    %     Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 );   

        if strcmp(Trimming_opt,'truncation')
            index = ( abs(Y).^2 <= alpha_y^2 * 1/delta );
            Ytr = abs(Y).^2.* index;  % column normalization
        elseif strcmp(Trimming_opt,'subset')
            Ytr = ( abs(Y).^2 >= alpha_y_subset^2 * 1/delta );  % column normalization
        elseif strcmp(Trimming_opt,'optimal')
            Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 );
        elseif strcmp(Trimming_opt,'optimal_Lu')
            Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + Lu_eps);    
        elseif strcmp(Trimming_opt,'proposed')
            Ytr = 1 - 1./ ( abs(Y).^2 * delta + delta - 2 );       
        end

        mu_min = 0;
        mu_max = 1/max(Ytr);
        counter = 0;
        if strcmp(Trimming_opt,'proposed')
            mu = 1;
            G = 1./(1-mu * Ytr);
        else
            while counter < 1e4
                counter = counter + 1;

                mu = (mu_min+mu_max)/2;
                G = 1./(1-mu * Ytr);
                f_tmp = (delta-1) * mean( abs(Z).^2.*G ) - mean(G);
                if f_tmp < 0
                    mu_min = mu;
                elseif f_tmp > 0
                    mu_max = mu;
                end

                if abs(f_tmp) < 1e-5 && abs(mu_max-mu_min)/mu_min < 1e-5
                    break;
                end

            end
        end

        % calculate rho_u
        D = mean(G);
        G2 = mean(abs(G).^2);
        G2Z2 = mean(abs(G).^2 .* abs(Z).^2);

        NUM = delta * D^2 - (delta - 1) * G2;
        DEN = (delta-1)^2 * G2Z2 - delta * D^2;

        rho_u2 = (delta-1)/delta * NUM/DEN;

        if rho_u2 < 0
            fprintf('below the phase transition\n');
            continue;
        end

        rho_x2 = rho_u2 * delta/(delta-1);
        Rho_x2(ddd,mmm) = rho_x2;
        alpha_x = sqrt(  rho_x2/(1+rho_x2) );
        Alpha_SE(ddd,mmm) = alpha_x;

        if SE_only == 'Y'
            continue;
        end

        %% simulations
        for nsim = 1:NSIM
        
            for rgb = 1:3, 
    
                fprintf('Color band %d\n', rgb)
                X = squeeze(X_full(:,:,rgb)); % Image x is n1 x n2 
                x = X(:);
                scaling = 1/norm(x) * sqrt(length(x));
                x = x * scaling;
                X = X * scaling;
                % affect the threshold

                for ll = 1:L
                    if strcmp(Mask_opt,'Y')
                        Masks(:,:,ll) = randsrc(n1,n2,[1i -1i 1 -1]);
                    else
                        Masks(:,:,ll) = randsrc(n1,n2,1);
                    end
                end


                % Make linear operators; 
                A = @(I) 1/sqrt(M) * fft2(conj(Masks) .* reshape(repmat(I,[1 L]), n1, n2, L));  % Input is n1 x n2 image, output is n1 x n2 x L array

            %     At = @(Y) sqrt(M) * sum(Masks .* ifft2(Y), 3) * size(Y,1) * size(Y,2);                       % Input is n1 x n2 x L array, output is n1 x n2 image
                At = @(Y) N/sqrt(M) * sum(Masks .* ifft2(Y), 3); 
                y = abs( A(X) );

                %% trim
                if strcmp(Trimming_opt,'truncation')
                    index = ( abs(y).^2 <= alpha_y^2 * 1/delta );
                    ytr = abs(y).^2.* index;  % column normalization
                elseif strcmp(Trimming_opt,'subset')
                    ytr = ( abs(y).^2 >= alpha_y_subset^2 * 1/delta );  % column normalization
                elseif strcmp(Trimming_opt,'optimal')
                    ytr = (abs(y).^2 * delta - 1) ./ ( abs(y).^2 * delta + sqrt(delta) - 1 );  
                elseif strcmp(Trimming_opt,'optimal_Lu')
                    ytr = 1 - 1./(abs(y).^2 * delta + Lu_eps); 
                elseif strcmp(Trimming_opt,'proposed')
                    ytr = 1 - 1./(abs(y).^2 * delta + delta - 2);       
                end

                % The power method
                z0 = randn(n1,n2) + 1i * randn(n1,n2);
                z0_new = z0;

                for tt = 1:Iteration_power,                                             % Truncated power iterations

                    z0_new = At( ( ytr.* A(z0) ) );

                    lambda = norm(z0_new(:))/norm(z0(:));


                    if strcmp(Trimming_opt,'optimal')
                        epsilon = 10;  % should use a large one for small delta !!
                    elseif strcmp(Trimming_opt,'proposed')
                        epsilon = 10;  % should use a large one for small delta !!
                    elseif  strcmp(Trimming_opt,'optimal_Lu') 
                        epsilon = 50;
                    else
                        epsilon = 0;
                    end
%                     epsilon = 0;

                    z0_new = z0_new + epsilon * z0;

                    x_power = sqrt(N) * z0_new(:)/norm(z0_new(:)); 

                    % measure
                    rho_sim = abs( x'*x_power/norm(x)/norm(x_power) );

                    z0_new = z0_new/norm(z0_new(:),'fro');  % no need to modify here. Normalize later
                    error_it = norm(z0(:)-z0_new(:))/norm(z0(:));
                    if ( error_it < 1e-10)
                        break;
                    end

                    z0 = reshape(z0_new,n1,n2);
                end

                Alpha_power(ddd,mmm) = Alpha_power(ddd,mmm) + rho_sim;
                Conv_error(ddd,mmm) = Conv_error(ddd,mmm) + error_it;
                
            
            end
            toc
            fprintf('**********************************************************************\n');
            t2 = clock;
     
            fprintf('elapsed time: %d secs\n',etime(t2,t1));
            fprintf('method = %s, delta = %.2f, nsim = %d\n',Trimming_opt, delta, nsim);
            fprintf('alpha = %f, alpha_SE = %f\n\n',Alpha_power(ddd,mmm)/nsim/3,Alpha_SE(ddd,mmm));
        end

        Alpha_power(ddd,mmm) = Alpha_power(ddd,mmm) / NSIM/3; % Three channels

    end

end
save('resultsCDP.mat','Alpha_SE','Alpha_power','delta_grid','Methods','SE_only');

FontSizeMy = 14;
LineWidthMy = 1.5;
Colors = {'r','g','b','m','c','k'} ;
Colors_SE = {'r^','g^','b^','m^','c^','k^'} ;
figure;
for mmm = 1:length(Methods)
    plot(delta_grid,Alpha_SE(:,mmm), Colors{mmm} ,'LineWidth',LineWidthMy);
    hold on; 
end
xlabel('delta'); 
ylabel('alpha');
legend(Methods);

if SE_only == 'N'
    hold on;
    for mmm = 1:length(Methods)
        plot(delta_grid,Alpha_power(:,mmm), Colors_SE{mmm} ,'LineWidth',LineWidthMy);
        hold on;
    end
end

set(gca,'fontsize',FontSizeMy);
legend(Methods,'fontsize',FontSizeMy);






