%**************************************************************************
% Spectral method: simulation & prediction
% By Junjie Ma
%**************************************************************************
tic
close all
clear all;
clc

randn('state',0);
rand('state',0);

namestr = 'venere' ;  % stanford, galaxy, columbia
stanstr = 'png';

Matrix_opt = 'doubleDFT'; % DFT, DFT_phase, doubldDFT, RandomCirculant
Phase_opt = 'Y';
SE_only = 'N';

X_full  = mat2gray(imread([namestr,'.',stanstr])) ;

X_full       = imresize(X_full, 0.1);

X_tmp = squeeze(X_full(:,:,1)); % Image x is n1 x n2 
x_tmp = X_tmp(:);
N = length(x_tmp);


% Methods = {'optimal_Lu','proposed','optimal','subset','truncation'};
Methods = {'subset'};

Lu_eps = 0.01;

alpha_y = 2;

alpha_y_subset = 1.5;

delta_grid = [2.1, 2.5:0.5:8];
% delta_grid = 4.5;
Iteration_power = 10000;

NSIM = 1;
  
Alpha_power = zeros(length(delta_grid),length(Methods));  
Conv_error = zeros(length(delta_grid),length(Methods));

Alpha_SE = zeros(length(delta_grid),length(Methods));
Rho_x2 = zeros(length(delta_grid),length(Methods));

t1 = clock;

for mmm = 1:length(Methods)
    
    Trimming_opt = Methods{mmm};
    
    for ddd = 1:length(delta_grid)

        delta = delta_grid(ddd);

        M = fix(delta * N);                        % number of measurements

        %% SE
        % search mu;

        flag = 1;
        MM = 1e7;
        Z = sqrt(1/delta) * (randn(MM,1)+1i*randn(MM,1))/sqrt(2);
        Y = abs(Z);
    %     Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 );   

        if strcmp(Trimming_opt,'truncation')
            index = ( abs(Y).^2 <= alpha_y^2 * 1/delta );
            Ytr = abs(Y).^2.* index;  % column normalization
        elseif strcmp(Trimming_opt,'subset')
            Ytr = ( abs(Y).^2 >= alpha_y_subset^2 * 1/delta );  % column normalization
        elseif strcmp(Trimming_opt,'optimal')
            Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 );
        elseif strcmp(Trimming_opt,'optimal_Lu')
            Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + Lu_eps);    
        elseif strcmp(Trimming_opt,'proposed')
            Ytr = 1 - 1./ ( abs(Y).^2 * delta + delta - 2 );       
        end

        mu_min = 0;
        mu_max = 1/max(Ytr);
        counter = 0;
        if strcmp(Trimming_opt,'proposed')
            mu = 1;
            G = 1./(1-mu * Ytr);
        else
            while counter < 1e4
                counter = counter + 1;

                mu = (mu_min+mu_max)/2;
                G = 1./(1-mu * Ytr);
                f_tmp = (delta-1) * mean( abs(Z).^2.*G ) - mean(G);
                if f_tmp < 0
                    mu_min = mu;
                elseif f_tmp > 0
                    mu_max = mu;
                end

                if abs(f_tmp) < 1e-5 && abs(mu_max-mu_min)/mu_min < 1e-5
                    break;
                end

            end
        end

        % calculate rho_u
        D = mean(G);
        G2 = mean(abs(G).^2);
        G2Z2 = mean(abs(G).^2 .* abs(Z).^2);

        NUM = delta * D^2 - (delta - 1) * G2;
        DEN = (delta-1)^2 * G2Z2 - delta * D^2;

        rho_u2 = (delta-1)/delta * NUM/DEN;

        if rho_u2 < 0
            fprintf('below the phase transition\n');
            continue;
        end

        rho_x2 = rho_u2 * delta/(delta-1);
        Rho_x2(ddd,mmm) = rho_x2;
        alpha_x = sqrt(  rho_x2/(1+rho_x2) );
        Alpha_SE(ddd,mmm) = alpha_x;

        if SE_only == 'Y'
            continue;
        end

        %% simulations
        for nsim = 1:NSIM
        
            for rgb = 1:3, 
    
                fprintf('Color band %d\n', rgb)
                X = squeeze(X_full(:,:,rgb)); % Image x is n1 x n2 
                x = X(:);
                x = x/norm(x) * sqrt(length(x)); % Be careful. If we don't do normalization, then, we need to carefuly tune epsilon in the power. Needs to be scaled
                
                
                takers = randperm(M,N);
                expand = sparse(takers,1:N,1,M,N);
                Phase = exp(-1i*rand(N,1)*2*pi);
                
                if strcmp(Matrix_opt,'DFT')
                    A = @(x) fft(expand*x) / sqrt(M);  % column normalization
                    At = @(x) (  expand'*ifft(x)* sqrt(M) );
                elseif strcmp(Matrix_opt,'DFT_phase')
                    A = @(x) fft(expand*(Phase.*x)) / sqrt(M);  % column normalization
                    At = @(x) conj(Phase) .* (  expand'*ifft(x)* sqrt(M) );
                elseif strcmp(Matrix_opt,'doubleDFT')
                    P_x = sign(randn(M,1));
                    P_y = sign(randn(M,1));
                    Phase1 = (P_x + 1i * P_y)/sqrt(2);

                    P_x = sign(randn(M,1));
                    P_y = sign(randn(M,1));
                    Phase2 = (P_x + 1i * P_y)/sqrt(2);
                    A = @(x) ifft( Phase2.* fft(Phase1 .* (expand*x) ) );  % column normalization
                    At = @(x) expand'* (  conj(Phase1) .* ifft( conj(Phase2) .* fft(x) ) );
                elseif strcmp(Matrix_opt,'RandomCirculant')
%                     P_x = sign(randn(M,1));
%                     P_y = sign(randn(M,1));
%                     Phase1 = (P_x + 1i * P_y)/sqrt(2);

                    P_x = sign(randn(M,1));
                    P_y = sign(randn(M,1));
                    Phase2 = (P_x + 1i * P_y)/sqrt(2);
                    A = @(x) ifft( Phase2.* fft( (expand*x) ) );  % column normalization
                    At = @(x) expand'* (   ifft( conj(Phase2) .* fft(x) ) );
                else
                    fprintf('no such option\n');
                    pause;
                end

                

                z_true = A(x);
                y = abs( z_true );

                %% trim

                if strcmp(Trimming_opt,'truncation')
                    index = ( abs(y).^2 <= alpha_y^2 * 1/delta );
                    ytr = abs(y).^2.* index;  % column normalization
                elseif strcmp(Trimming_opt,'subset')
                    ytr = ( abs(y).^2 >= alpha_y_subset^2 * 1/delta );  % column normalization
                elseif strcmp(Trimming_opt,'optimal')
                    ytr = (abs(y).^2 * delta - 1) ./ ( abs(y).^2 * delta + sqrt(delta) - 1 );  
                elseif strcmp(Trimming_opt,'optimal_Lu')
                    ytr = 1 - 1./(abs(y).^2 * delta + Lu_eps); 
                elseif strcmp(Trimming_opt,'proposed')
                    ytr = 1 - 1./(abs(y).^2 * delta + delta - 2);       
                end

                % The power method
                z0 = randn(N,1) + 1i * randn(N,1);

                normest = sqrt(N);
                z0_new = z0;

                for tt = 1:Iteration_power,                                             % Truncated power iterations

                    z0_new = At( ( ytr.* A(z0) ) );

                    lambda = norm(z0_new(:))/norm(z0(:));

                    z0_new = z0_new(:);

                    if strcmp(Trimming_opt,'optimal')
                        epsilon = 10;  % should use a large one for small delta !!
                    elseif strcmp(Trimming_opt,'proposed')
                        epsilon = 10;  % should use a large one for small delta !!
                    elseif  strcmp(Trimming_opt,'optimal_Lu') 
                        epsilon = 50;
                    else
                        epsilon = 0;
                    end
%                     epsilon = 0;

                    z0_new = z0_new + epsilon * z0;

                    x_power = normest * z0_new/norm(z0_new); 

                    % measure
                    rho_sim = abs( x'*x_power/norm(x)/norm(x_power) );

                    z0_new = z0_new/norm(z0_new,'fro');  % no need to modify here. Normalize later
                    error_it = norm(z0-z0_new)/norm(z0);
                    if ( error_it < 1e-10)
                        break;
                    end

                    z0 = z0_new;
                end

                Alpha_power(ddd,mmm) = Alpha_power(ddd,mmm) + rho_sim;
                Conv_error(ddd,mmm) = Conv_error(ddd,mmm) + error_it;
            
            end
            toc
            fprintf('**********************************************************************\n');
            t2 = clock;
     
            fprintf('elapsed time: %d secs\n',etime(t2,t1));
            fprintf('method = %s, delta = %.2f, nsim = %d\n',Trimming_opt, delta, nsim);
            fprintf('alpha = %f, alpha_SE = %f\n\n',Alpha_power(ddd,mmm)/nsim/3,Alpha_SE(ddd,mmm));
        end

        Alpha_power(ddd,mmm) = Alpha_power(ddd,mmm) / NSIM/3; % Three channels

    end

end
save('results.mat','Alpha_SE','Alpha_power','delta_grid','Methods','SE_only');

FontSizeMy = 14;
LineWidthMy = 1.5;
% Colors = {'r','g','b','m','c','k'} ;
% Colors_SE = {'ro','go','bo','mo','co','ko'} ; original
Colors_SE = {'r','g','b','m','c','k'} ; % modified on 26 Feb 2019
Colors = {'r+','g+','b+','m+','c+','k+'} ;
figure;
for mmm = 1:length(Methods)
    plot(delta_grid,Alpha_SE(:,mmm), Colors{mmm} ,'LineWidth',LineWidthMy);
    hold on; 
end
xlabel('delta'); 
ylabel('alpha');
legend(Methods);

if SE_only == 'N'
    hold on;
    for mmm = 1:length(Methods)
        plot(delta_grid,Alpha_power(:,mmm), Colors_SE{mmm} ,'LineWidth',LineWidthMy);
        hold on;
    end
end

set(gca,'fontsize',FontSizeMy);
legend(Methods,'fontsize',FontSizeMy);






