%**************************************************************************
% Accuracy of SE
% By Junjie Ma
%**************************************************************************
clc
close all
clear all;
clc

it_start = 1;
N = 30000;
alpha_y = 1.5;

L = 3;                                    % It is set to a very large number to see the agreement with evolution. You can try a smaller one.
delta = L;

Iteration = 50;

M = fix(L * N);                        % number of measurements

NSIM = 10;

% randn('state',0);
% rand('state',0);

SE_only = 'N';
Trimming_opt = 'optimal_Lu';
Phase_random = 'Y';

% alpha and sigma for u
alpha_0 = 0.2;
sigma2_0 = 1;

%% derive mu
flag = 1;
MM = 2e7;
Z = sqrt(1/delta) * (randn(MM,1)+1i*randn(MM,1))/sqrt(2);
Y = abs(Z);
%     Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 );   

if strcmp(Trimming_opt,'truncation')
    index = ( abs(Y).^2 <= alpha_y^2 * 1/delta );
    Ytr = abs(Y).^2.* index;  % column normalization
    Y_max_true = max(Ytr);
elseif strcmp(Trimming_opt,'subset')
    Ytr = ( abs(Y).^2 >= alpha_y^2 * 1/delta );  % column normalization
    Y_max_true = 1;
elseif strcmp(Trimming_opt,'MM')
    Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 );
    Y_max_true = 1;
elseif strcmp(Trimming_opt,'optimal_Lu')
    Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta);    
    Y_max_true = 1;
elseif strcmp(Trimming_opt,'proposed')
    Ytr = 1 - 1./ ( abs(Y).^2 * delta + delta - 2 );  
    Y_max_true = 1;
end

mu_min = 0;
mu_max = Y_max_true;
if strcmp(Trimming_opt,'proposed')
    mu = 1;
else
    counter = 0;
    while counter < 1e4
        counter = counter + 1;

        mu = (mu_min+mu_max)/2;
        G = 1./(1-mu * Ytr);
        f_tmp = mean( delta * abs(Z).^2.*G )/ mean(G) - delta/(delta-1);
        if f_tmp < 0
            mu_min = mu;
        elseif f_tmp > 0
            mu_max = mu;
        end

        if abs(f_tmp) < 1e-7 && abs(mu_max-mu_min)/mu_min < 1e-7
            break;
        end

    end
end
%%

Alpha_SE = zeros(Iteration,1);
Sigma2_SE = zeros(Iteration,1);
Cov_w_tilde_unnorm_SE = zeros(Iteration,1);
Cov_w_tilde_SE = zeros(Iteration,1);

Alpha_sim = zeros(Iteration,NSIM);
Sigma2_sim = zeros(Iteration,NSIM);
Cov_w_tilde_sim = zeros(Iteration,NSIM);

Rho_simulation = zeros(Iteration, NSIM);



%% simulation
if SE_only ~= 'Y'
    
    for nsim = 1:NSIM
    
        x = (randn(N,1) + 1i * randn(N,1))/sqrt(2);

        takers = randperm(M,N); % indices for the nonzero entries.
        expand = sparse(takers,1:N,1,M,N);

        Phase = exp(1i*rand(N,1)*2*pi);

        if Phase_random == 'N'
            A = @(x) fft(expand*x) / sqrt(M);  % column normalization
            At = @(x) expand'*ifft(x)* sqrt(M);
        else
            A = @(x) fft(expand*(Phase.*x)) / sqrt(M);  % column normalization
            At = @(x) conj(Phase) .* ( expand'*ifft(x)* sqrt(M) );
        end

        z_true = A(x);
        y = abs( z_true );

        if strcmp(Trimming_opt,'truncation')
            index = ( abs(y).^2 <= alpha_y^2 * 1/delta );
            ytr = abs(y).^2.* index;  % column normalization
        elseif strcmp(Trimming_opt,'subset')
            ytr = ( abs(y).^2 >= alpha_y^2 * 1/delta );  % column normalization
        elseif strcmp(Trimming_opt,'MM')
            ytr = (abs(y).^2 * delta - 1) ./ ( abs(y).^2 * delta + sqrt(delta) - 1 );  
        elseif strcmp(Trimming_opt,'optimal_Lu')
            ytr = 1 - 1./(abs(y).^2 * delta + 0); 
        elseif strcmp(Trimming_opt,'proposed')
            ytr = 1 - 1./(abs(y).^2 * delta + delta - 2);       
        end

        %% Simulations

        w_tilde = sqrt(sigma2_0/delta) * (randn(M,1) + 1i * randn(M,1)) / sqrt(2);     % w_tilde inclue sigma

        z_hat = alpha_0 * z_true + w_tilde;  % this is our definition of 

        C_seq = zeros(Iteration,1);

        Test_sim = zeros(Iteration,1);

        %  Note that G, and div are fixed as long as mu is fixed
        G = 1./(1 - mu * ytr); 
        div = mean(G);  

        z_hat_old = ones(M,1);

        for it = 1:Iteration

            p_hat = 1/div * ( G  - div ) .* z_hat ; % the scaling 1/div does not matter?

            z_hat = ( delta * A(At(p_hat)) - p_hat );

            % measure the quantities
            alpha_sim = z_true'*z_hat/norm(z_true)^2;
            sigma2_sim = norm(z_hat-alpha_sim*z_true)^2/norm(z_true)^2;
            w_tilde_new = z_hat - alpha_sim * z_true;

            Alpha_sim(it,nsim) = abs(alpha_sim);
            Sigma2_sim(it,nsim) = sigma2_sim;
            Cov_w_tilde_sim(it,nsim) = abs(w_tilde'*w_tilde_new)/norm(w_tilde_new)/norm(w_tilde);

            x_hat = At(z_hat);
            Rho_simulation(it,nsim) = abs(x'*x_hat)/norm(x)/norm(x_hat);

            % update
            w_tilde = w_tilde_new;

        end 
    end
    
end

%% SE
alpha = alpha_0;
sigma2 = sigma2_0;

NN = 5e7;

z = sqrt(1/delta) * ( randn(NN,1) + 1i * randn(NN,1) )/sqrt(2);
Y = abs(z);

% Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 ); 
if strcmp(Trimming_opt,'truncation')
    index = ( abs(Y).^2 <= alpha_y^2 * 1/delta );
    Ytr = abs(Y).^2.* index;  % column normalization
elseif strcmp(Trimming_opt,'subset')
    Ytr = ( abs(Y).^2 >= alpha_y^2 * 1/delta );  % column normalization
elseif strcmp(Trimming_opt,'MM')
    Ytr = (abs(Y).^2 * delta - 1) ./ ( abs(Y).^2 * delta + sqrt(delta) - 1 );  
elseif strcmp(Trimming_opt,'optimal_Lu')
    Ytr = 1 - 1./(abs(Y).^2 * delta +0); 
elseif strcmp(Trimming_opt,'proposed')
    Ytr = 1 - 1./(abs(Y).^2 * delta + delta - 2);       
end

Test_SE = zeros(Iteration,1);

GG = 1./(1 - mu * Ytr);
D = mean(GG);
G_tilde = GG - D;
div_p = mean(GG);

% psi functions defined according to paper
if strcmp(Trimming_opt,'proposed')
    psi1 = delta/(delta-1);
    psi2 = 1+1/(delta-1)^2;
    psi3 = sqrt(delta^2+2)/(delta-1);
else
    z_tilde = sqrt(delta) * z;
    psi1 = mean( abs(z_tilde).^2 .* GG)/D;
    psi2 = mean( GG.^2 ) / D^2;
    psi3 = sqrt( mean(abs(z_tilde).^2 .* GG.^2) ) / D;
end

alpha = alpha_0;
sigma2 = sigma2_0;

% assumption on p_hat = alpha * z + sigma * CN(0,1);  NOT CN(0,1/delta)
for it = 1:Iteration

    % a different approach
    alpha_new = (delta-1)*alpha*(psi1-1);
    sigma2_new = (delta-1) * ( alpha^2 * (psi3^2-psi1^2) + (psi2-1) * sigma2 );

    Alpha_SE(it) = alpha_new;
    Sigma2_SE(it) = sigma2_new;
    
    if it==1
        Cov_w_tilde_unnorm_SE(it) = 0;
    else
        if it==2
            alpha_previous = alpha_0;
        else
            alpha_previous = Alpha_SE(it-2);
        end
        
        sigma = sqrt(sigma2);
        sigma_new = sqrt(sigma2_new);
        
        Cov_w_tilde_unnorm_SE(it) = (delta-1)*( alpha*alpha_previous/delta * ( psi3^2-psi1^2 ) + ...
            (psi2-1) * Cov_w_tilde_unnorm_SE(it-1) );
        
        % normalization: E[w^2]=1/delta * sigma2
        w_tilde_norm = sqrt(1/delta * sigma2 );
        w_tilde_new_norm = sqrt(1/delta * sigma2_new );
        Cov_w_tilde_SE(it) = Cov_w_tilde_unnorm_SE(it)/w_tilde_norm/w_tilde_new_norm;
    end
    
    alpha = alpha_new;
    sigma2 = sigma2_new;
end


% figure;
% plot(1:Iteration,abs(Alpha_sim)./sqrt(Sigma2_sim),'r',1:Iteration,Alpha_SE./sqrt(Sigma2_SE),'b.');
% xlabel('iteration');
% legend('alpha/sigma');
% 
% 
% figure;
% plot(1:Iteration,Cov_w_tilde_sim,'r',1:Iteration,Cov_w_tilde_SE,'b.')
% xlabel('iteration');
% ylabel('correlation-normalized');

%================= with errorbars =======
% alpha & sigma2
FontSizeMy = 16;
LineWidthMy = 1.5;

Alpha_sim_mean = mean(Alpha_sim,2);   % 0 use normaization N-1, 1 use N
Alpha_sim_std = std(Alpha_sim,0,2);   % 0 use normaization N-1, 1 use N

Sigma2_sim_mean = mean(Sigma2_sim,2);   % 0 use normaization N-1, 1 use N
Sigma2_sim_std = std(Sigma2_sim,0,2);   % 0 use normaization N-1, 1 use N

figure;
plot(1:Iteration,Alpha_SE,'b.');
hold on;
errorbar(1:Iteration,Alpha_sim_mean,Alpha_sim_std,'r','LineWidth',LineWidthMy);
xlabel('iteration');
ylabel('$\alpha$');

figure;
plot(1:Iteration,Sigma2_SE,'b.');
hold on;
errorbar(1:Iteration,Sigma2_sim_mean,Sigma2_sim_std,'r','LineWidth',LineWidthMy);
xlabel('iteration');
ylabel('$\sigma^2$');


% cosine similarity
Rho_simulation_mean = mean(Rho_simulation,2);   % 0 use normaization N-1, 1 use N
Rho_simulation_std = std(Rho_simulation,0,2);   % 0 use normaization N-1, 1 use N

Rho_SE = Alpha_SE./sqrt( Sigma2_SE * (delta-1)/delta + Alpha_SE.^2 );


figure;
plot(1:Iteration,Rho_SE,'b.');
hold on;
errorbar(1:Iteration,Rho_simulation_mean,Rho_simulation_std,'r','LineWidth',LineWidthMy);
xlabel('iteration');
ylabel('$\rho(\mathbf{x}_\star,\hat{\mathbf{x}})$');

Cov_w_simulation_mean = mean(Cov_w_tilde_sim,2);   % 0 use normaization N-1, 1 use N
Cov_w_simulation_std = std(Cov_w_tilde_sim,0,2);   % 0 use normaization N-1, 1 use N

figure;
plot(1:Iteration,Cov_w_tilde_SE,'b.');
hold on;
errorbar(1:Iteration,Cov_w_simulation_mean,Cov_w_simulation_std,'r','LineWidth',LineWidthMy);
xlabel('iteration');
ylabel('$\rho(w^t,w^{t+1})$');

